package verifier;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeVerifier;
import meerkat.protobuf.ConcreteCrypto.ElGamalCiphertext;
import meerkat.protobuf.Crypto;
import meerkat.protobuf.Mixing;
import org.bouncycastle.math.ec.ECPoint;
import prover.ProofOrganizer;
import qilin.primitives.RandomOracle;
import qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;


public class Verifier implements Mix2ZeroKnowledgeVerifier {


    /**
     * Return true iff the proof is valid.
     * @param in1
     * @param in2
     * @param out1
     * @param out2
     * @return
     */

    ECGroup group;
    RandomOracle randomOracle;
    ECElGamalEncryption encryptor;
    ECPoint g,h;

    public Verifier(ECElGamalEncryption encryptor, RandomOracle randomOracle) {
        this.encryptor = encryptor;
        this.group = encryptor.getGroup();
        this.g = group.getGenerator();
        this.h = encryptor.getElGamalPK().getPK();
        this.randomOracle = randomOracle;
    }

    public BigInteger hash(Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle input) {
        byte[] arr = input.toByteArray();
        return new BigInteger(this.randomOracle.hash(arr,arr.length));
    }

    private ECPoint convert2ECPoint(ByteString bs){
        return group.decode(bs.toByteArray());
    }


    public boolean verify(Crypto.RerandomizableEncryptedMessage in1,
                          Crypto.RerandomizableEncryptedMessage in2,
                          Crypto.RerandomizableEncryptedMessage out1,
                          Crypto.RerandomizableEncryptedMessage out2,
                          Mixing.ZeroKnowledgeProof proof) throws InvalidProtocolBufferException {

        ProofOrganizer organizer = new ProofOrganizer(in1,in2,out1,out2);
        return verifyOrProof(organizer.getOrProofInput(ProofOrganizer.OrProofOrder.first), proof.getFirst())&&
                verifyOrProof(organizer.getOrProofInput(ProofOrganizer.OrProofOrder.second), proof.getSecond())&&
                verifyOrProof(organizer.getOrProofInput(ProofOrganizer.OrProofOrder.third), proof.getThird())&&
                verifyOrProof(organizer.getOrProofInput(ProofOrganizer.OrProofOrder.fourth), proof.getFourth());

    }

    public boolean verifyOrProof(ProofOrganizer.OrProofInput orProofInput, Mixing.ZeroKnowledgeProof.OrProof orProof)
            throws InvalidProtocolBufferException {
        ElGamalCiphertext e1ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(orProofInput.e1);
        ElGamalCiphertext e2ElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(orProofInput.e2);
        ElGamalCiphertext e1TagElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(orProofInput.e1New);
        ElGamalCiphertext e2TagElGamal = ECElGamalEncryption.rerandomizableEncryptedMessage2ElGamalCiphertext(orProofInput.e2New);
        return verifyElGamaOrProof(e1ElGamal,e2ElGamal,e1TagElGamal,e2TagElGamal,orProof);
    }



    private ECPoint g1,g2,h1,h2;
    private ECPoint g1Tag,g2Tag,h1Tag,h2Tag;
    private ECPoint u,v,uTag,vTag;
    private BigInteger c1,c2,z,zTag;
    private  Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle forRandomOracle;

    private void parseOrProof(Mixing.ZeroKnowledgeProof.OrProof orProof){
        forRandomOracle =
                Mixing.ZeroKnowledgeProof.OrProof.ForRandomOracle.newBuilder()
                        .setG1(orProof.getG1())
                        .setH1(orProof.getH1())
                        .setG2(orProof.getG2())
                        .setH2(orProof.getH2())
                        .setG1Tag(orProof.getG1Tag())
                        .setH1Tag(orProof.getH1Tag())
                        .setG2Tag(orProof.getG2Tag())
                        .setH2Tag(orProof.getH2Tag())
                        .setU(orProof.getU())
                        .setV(orProof.getV())
                        .setUTag(orProof.getUTag())
                        .setVTag(orProof.getVTag())
                        .build();
        g1 = convert2ECPoint(orProof.getG1());
        g2 = convert2ECPoint(orProof.getG2());
        h1 = convert2ECPoint(orProof.getH1());
        h2 = convert2ECPoint(orProof.getH2());
        g1Tag = convert2ECPoint(orProof.getG1Tag());
        g2Tag = convert2ECPoint(orProof.getG2Tag());
        h1Tag = convert2ECPoint(orProof.getH1Tag());
        h2Tag = convert2ECPoint(orProof.getH2Tag());
        u = convert2ECPoint(orProof.getU());
        v = convert2ECPoint(orProof.getV());
        uTag = convert2ECPoint(orProof.getUTag());
        vTag = convert2ECPoint(orProof.getVTag());
        c1 = new BigInteger(orProof.getC1().toByteArray());
        c2 = new BigInteger(orProof.getC2().toByteArray());
        z = new BigInteger(orProof.getZ().toByteArray());
        zTag = new BigInteger(orProof.getZTag().toByteArray());
    }

    private List<Condition> createConditionsList(ElGamalCiphertext e1,
                                                 ElGamalCiphertext e2,
                                                 ElGamalCiphertext e1New,
                                                 ElGamalCiphertext e2New){
        List<Condition> conditions = new ArrayList<Condition>();
        conditions.add(new Condition<ECPoint>( g1,g,"g1 != g"));
        conditions.add(new Condition<ECPoint>( h1,group.add(convert2ECPoint(e1New.getC1()),
                group.negate(convert2ECPoint(e1.getC1()))),"h1 != e1New.c1/e1.c1"));
        conditions.add(new Condition<ECPoint>( g2,h,"g2 != h"));
        conditions.add(new Condition<ECPoint>( h2,group.add(convert2ECPoint(e1New.getC2()),
                group.negate(convert2ECPoint(e1.getC2()))),"h2 != e1New.c2/e1.c2"));
        conditions.add(new Condition<ECPoint>( g1Tag,g,"g1Tag != g"));
        conditions.add(new Condition<ECPoint>( h1Tag,group.add(convert2ECPoint(e2New.getC1()),
                group.negate(convert2ECPoint(e2.getC1()))),"h1Tag != e2New.c1/e2.c1"));
        conditions.add(new Condition<ECPoint>( g2Tag,h,"g2Tag != h"));
        conditions.add(new Condition<ECPoint>( h2Tag,group.add(convert2ECPoint(e2New.getC2()),
                group.negate(convert2ECPoint(e2.getC2()))),"h2Tag != e2New.c2/e2.c2"));
        conditions.add(new Condition<BigInteger>(c1.add(c2).mod(group.orderUpperBound()),
                hash(forRandomOracle).mod(group.orderUpperBound()).mod(group.orderUpperBound()),
                "(c1 + c2 ) % group size == hash (imput + step1) % group size"));
        conditions.add(new Condition<ECPoint>( group.multiply(g1, z),
                group.add(u, group.multiply(h1,c1)),"g1 ^ z != u * ( h1 ^ c1 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(g2, z),
                group.add(v, group.multiply(h2,c1)),"g2 ^ z != v * ( h2 ^ c1 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(g1Tag, zTag),
                group.add(uTag, group.multiply(h1Tag,c2)),"g1Tag ^ zTag != uTag * ( h1Tag ^ c2 )"));
        conditions.add(new Condition<ECPoint>( group.multiply(g2Tag, zTag),
                group.add(vTag, group.multiply(h2Tag,c2)),"g2Tag ^ zTag != vTag * ( h2Tag ^ c2 )"));
        return conditions;
    }

    public boolean verifyElGamaOrProof(ElGamalCiphertext e1,
                                 ElGamalCiphertext e2,
                                 ElGamalCiphertext e1New,
                                 ElGamalCiphertext e2New,
                                 Mixing.ZeroKnowledgeProof.OrProof orProof)
    {
        parseOrProof(orProof);
        List<Condition> conditions = createConditionsList(e1,e2,e1New,e2New);

        boolean result = true;
        for (Condition condition: conditions) {
            if(!condition.test()){
                if (result){
                    result = false;
                    System.out.print("\n\n\n");
                }
                System.out.println(condition.errorDescripton);
            }
        }
        return result;
    }

    private class Condition<T>{

        T given, expected;
        String errorDescripton;

        Condition(T given,T expected,String descripton){
            this.given = given;
            this.errorDescripton = descripton;
            this.expected = expected;
        }

        boolean test(){
            return given.equals(expected);
        }
    }
}
